"""
Description: This file contains the class FiwareConnections,
which is used to store the connection parameters for the Fiware and CrateDB connections.
Author: Martin Altenburger
"""

from asyncio import sleep
import os
from datetime import datetime, timedelta, timezone
from typing import Union
import concurrent.futures
import multiprocessing
from loguru import logger
import numpy as np
import pandas as pd
import requests
from dateutil import tz
from filip.clients.ngsi_v2 import ContextBrokerClient
from filip.models.base import DataType, FiwareHeaderSecure
from filip.models.ngsi_v2.base import NamedMetadata
from filip.models.ngsi_v2.context import (
    ContextAttribute,
    ContextEntity,
    NamedCommand,
    NamedContextAttribute,
)
from encodapy.config import (
    AttributeModel,
    AttributeTypes,
    CommandModel,
    DataQueryTypes,
    DefaultEnvVariables,
    InputModel,
    OutputModel,
    TimerangeTypes,
    ConfigModel,
)
from encodapy.utils.error_handling import NoCredentials, InterfaceNotActive
from encodapy.utils.cratedb import CrateDBConnection
from encodapy.utils.fiware_auth import BaererToken
from encodapy.utils.models import (
    InputDataAttributeModel,
    InputDataEntityModel,
    MetaDataModel,
    OutputDataAttributeModel,
    OutputDataEntityModel,
    FiwareDatapointParameter,
    FiwareAuth,
    FiwareParameter,
    DatabaseParameter,
    FiwareConnectionParameter,
)
from encodapy.utils.units import (
    DataUnits,
    get_time_unit_seconds,
    get_unit_adjustment_factor,
)


class FiwareConnection:
    """
    Class for the connection to the Fiware plattform.
    Only a helper class.
    """

    def __init__(self):

        self.fiware_conn_params: FiwareConnectionParameter = None
        self.fiware_token_client: BaererToken = None
        self.fiware_header: FiwareHeaderSecure = None
        self.cb_client: ContextBrokerClient = None
        self.crate_db_client: CrateDBConnection = None
        self.config: ConfigModel

    def load_fiware_params(self):
        """
        Load the Fiware connection parameters.
        """

        if os.getenv(
            "FIWARE_AUTH", str(DefaultEnvVariables.FIWARE_AUTH.value)
        ).lower() in ("true", "1", "t"):

            if (
                os.environ.get("FIWARE_CLIENT_ID") is not None
                and os.environ.get("FIWARE_CLIENT_PW") is not None
                and os.environ.get("FIWARE_TOKEN_URL") is not None
            ):
                fiware_auth = FiwareAuth(
                    client_id=os.environ.get("FIWARE_CLIENT_ID"),
                    client_secret=os.environ.get("FIWARE_CLIENT_PW"),
                    token_url=os.environ.get("FIWARE_TOKEN_URL"),
                )
            elif os.environ.get("FIWARE_BAERER_TOKEN") is not None:
                fiware_auth = FiwareAuth(
                    baerer_token=os.environ.get("FIWARE_BAERER_TOKEN")
                )
            else:
                logger.error("No authentication credentials available")
                raise NoCredentials
        else:
            fiware_auth = None

        fiware_params = FiwareParameter(
            cb_url=os.environ.get("CB_URL", DefaultEnvVariables.CB_URL.value),
            service=os.environ.get(
                "FIWARE_SERVICE", DefaultEnvVariables.FIWARE_SERVICE.value
            ),
            service_path=os.environ.get(
                "FIWARE_SERVICE_PATH", DefaultEnvVariables.FIWARE_SERVICE_PATH.value
            ),
            authentication=fiware_auth,
        )

        database_params = DatabaseParameter(
            crate_db_url=os.environ.get(
                "CRATE_DB_URL", DefaultEnvVariables.CRATE_DB_URL.value
            ),
            crate_db_user=os.environ.get(
                "CRATE_DB_USER", DefaultEnvVariables.CRATE_DB_USER.value
            ),
            crate_db_pw=os.environ.get(
                "CRATE_DB_PW", DefaultEnvVariables.CRATE_DB_PW.value
            ),
            crate_db_ssl=os.environ.get(
                "CRATE_DB_SSL", str(DefaultEnvVariables.CRATE_DB_SSL.value)
            ).lower()
            in ("true", "1", "t"),
        )

        self.fiware_conn_params = FiwareConnectionParameter(
            fiware_params=fiware_params, database_params=database_params
        )

    def check_fiware_connection(self) -> None:
        """
        Check the Fiware connection. Are there any entities available?
        """
        if self.cb_client is None:
            raise InterfaceNotActive("ContextBrokerClient is not active")

        if len(self.cb_client.get_entity_list()) == 0:
            logger.error(
                "No entities were found in the ContextBrokerClient. "
                "Ensure that the configuration for the FIWARE entities is valid."
            )

    def prepare_fiware_connection(self):
        """
        Prepare the Fiware connection.
        """
        fiware_auth = self.fiware_conn_params.fiware_params.authentication

        if fiware_auth is not None:

            if fiware_auth.baerer_token is not None:
                self.fiware_token_client = BaererToken(token=fiware_auth.baerer_token)
            else:
                self.fiware_token_client = BaererToken(
                    client_id=fiware_auth.client_id,
                    client_secret=fiware_auth.client_secret,
                    token_url=fiware_auth.token_url,
                )
            self.fiware_header = FiwareHeaderSecure(
                service=self.fiware_conn_params.fiware_params.service,
                service_path=self.fiware_conn_params.fiware_params.service_path,
                authorization=self.fiware_token_client.baerer_token,
            )
        else:
            self.fiware_header = FiwareHeaderSecure(
                service=self.fiware_conn_params.fiware_params.service,
                service_path=self.fiware_conn_params.fiware_params.service_path,
            )

        self.cb_client = ContextBrokerClient(
            url=self.fiware_conn_params.fiware_params.cb_url,
            fiware_header=self.fiware_header,
        )
        self.check_fiware_connection()

        self.crate_db_client = CrateDBConnection(
            crate_db_url=self.fiware_conn_params.database_params.crate_db_url,
            crate_db_user=self.fiware_conn_params.database_params.crate_db_user,
            crate_db_pw=self.fiware_conn_params.database_params.crate_db_pw,
            crate_db_ssl=self.fiware_conn_params.database_params.crate_db_ssl,
        )

    def update_authentication(self):
        """
        Update the authentication.
        """
        if self.fiware_conn_params.fiware_params.authentication is not None and (
            self.fiware_token_client.check_token() is False
        ):
            self.fiware_header.__dict__["authorization"] = (
                self.fiware_token_client.baerer_token
            )

    def _get_last_timestamp_for_fiware_output(
        self, output_entity: OutputModel
    ) -> tuple[OutputDataEntityModel, Union[datetime, None]]:
        """
        Function to get the latest timestamps of the output entity from the FIWARE platform

        Args:
            output_entity (OutputModel): Output entity

        Returns:
            tuple[OutputDataEntityModel, Union[datetime, None]]:
                - OutputDataEntityModel with timestamps for the attributes
                - the latest timestamp of the output entity for the attribute
                with the oldest value (None if no timestamp is available)
        """

        output_attributes_entity = self.cb_client.get_entity_attributes(
            entity_id=output_entity.id_interface
        )

        output_attributes_controller = {
            item.id_interface: item.id for item in output_entity.attributes
        }

        timestamps = []
        for attr in list(output_attributes_entity.keys()):
            if attr not in list(output_attributes_controller.keys()):
                continue
            if (
                output_attributes_entity[attr].metadata.get("TimeInstant") is not None
                and output_attributes_entity[attr].metadata.get("TimeInstant").value
                is not None
            ):
                timestamps.append(
                    OutputDataAttributeModel(
                        id=output_attributes_controller[attr],
                        latest_timestamp_output=datetime.strptime(
                            output_attributes_entity[attr]
                            .metadata.get("TimeInstant")
                            .value,
                            "%Y-%m-%dT%H:%M:%S.%f%z",
                        ),
                    )
                )

        if len(timestamps) > 0:
            timestamp_latest_output = min(
                item.latest_timestamp_output for item in timestamps
            )
        else:
            timestamp_latest_output = None

        return (
            OutputDataEntityModel(id=output_entity.id, attributes_status=timestamps),
            timestamp_latest_output,
        )

    def _get_metadata_from_fiware(
        self, fiware_attribute: ContextAttribute
    ) -> MetaDataModel:
        """Function to get the metadata from the fiware attribute

        Args:
            fiware_attribute (ContextAttribute): Fiware attribute

        Returns:
            MetaDataModel: Model with the metadata (timestamp, unit) of the attribute if available
        """
        metadata_lowercase = {
            k.lower(): v for k, v in fiware_attribute.metadata.items()
        }

        metadata_model = MetaDataModel()

        if metadata_lowercase.get("timeinstant") is not None:
            metadata_model.timestamp = datetime.strptime(
                metadata_lowercase.get("timeinstant").value, "%Y-%m-%dT%H:%M:%S.%f%z"
            )

        try:
            if metadata_lowercase.get("unitcode") is not None:
                metadata_model.unit = DataUnits(
                    metadata_lowercase.get("unitcode").value
                )
            elif metadata_lowercase.get("unittext") is not None:
                metadata_model.unit = DataUnits(
                    metadata_lowercase.get("unittext").value
                )
            elif metadata_lowercase.get("unit") is not None:
                metadata_model.unit = DataUnits(metadata_lowercase.get("unit").value)
        except ValueError as err:
            logger.error(
                f"Unit code {metadata_lowercase.get('unitcode').value} not available: {err}"
            )

        return metadata_model

    def get_data_from_fiware(
        self,
        method: DataQueryTypes,
        entity: InputModel,
        timestamp_latest_output: Union[datetime, None],
    ) -> Union[InputDataEntityModel, None]:
        """
        Function fetches the data for evaluation which have not yet been evaluated.
            First get the last timestamp of the output entity. Then get the data from
            the entity since the last timestamp of the output entity from cratedb.
        Args:
            - method (DataQueryTypes): Keyword for type of query
            - entity (InputModel): Input entity
            - timestamp_latest_output (datetime): Timestamp of the last output

        Returns:
            - InputDataEntityModel: Model with the input data or None
            if the connection to the platform is not available

        """

        attributes_timeseries = {}
        attributes_values = []

        if self.cb_client is None:
            raise InterfaceNotActive
        try:

            fiware_input_entity_type = self.cb_client.get_entity(
                entity_id=entity.id_interface
            ).type
            fiware_input_entity_attributes = self.cb_client.get_entity_attributes(
                entity_id=entity.id_interface, entity_type=fiware_input_entity_type
            )
        except requests.exceptions.ConnectionError as err:
            logger.error(f"""No connection to platform (ConnectionError): {err}""")

            return None

        for attribute in entity.attributes:

            if attribute.id_interface not in fiware_input_entity_attributes:
                logger.error(
                    f"Attribute {attribute.id_interface} not found in entity {entity.id_interface}"
                )
                attributes_values.append(
                    InputDataAttributeModel(
                        id=attribute.id,
                        data=None,
                        data_type=attribute.type,
                        data_available=False,
                        latest_timestamp_input=None,
                        unit=None,
                    )
                )

                continue

            if attribute.type == AttributeTypes.TIMESERIES:
                attributes_timeseries[attribute.id] = {
                    "id_interface": attribute.id_interface,
                    "metadata": self._get_metadata_from_fiware(
                        fiware_input_entity_attributes[attribute.id_interface]
                    ),
                }

            elif attribute.type == AttributeTypes.VALUE:

                metadata = self._get_metadata_from_fiware(
                    fiware_input_entity_attributes[attribute.id_interface]
                )
                attributes_values.append(
                    InputDataAttributeModel(
                        id=attribute.id,
                        data=fiware_input_entity_attributes[
                            attribute.id_interface
                        ].value,
                        data_type=AttributeTypes.VALUE,
                        data_available=(
                            fiware_input_entity_attributes[attribute.id_interface].value
                            is not None
                        ),
                        latest_timestamp_input=metadata.timestamp,
                        unit=metadata.unit,
                    )
                )
            else:
                logger.warning(
                    f"Attribute type {attribute.type} for attribute {attribute.id}"
                    f"of entity {entity.id} not supported."
                )
                raise NotImplementedError

        if len(attributes_timeseries) > 0:

            attributes_values.extend(
                self.get_data_from_datebase(
                    entity=ContextEntity(
                        id=entity.id_interface, type=fiware_input_entity_type
                    ),
                    entity_attributes=attributes_timeseries,
                    method=method,
                    timestamp_latest_output=timestamp_latest_output,
                )
            )

        return InputDataEntityModel(id=entity.id, attributes=attributes_values)

    def _calculate_timerange(
        self,
        time_now: datetime,
        last_timestamp: Union[datetime, None],
        timerange_value: int,
        timerange_type: Union[TimerangeTypes, None],
    ) -> tuple[str, str]:
        """Function to calculate the timerange for the input data query based on
        a fixed timerange from the configuration

        Args:
            time_now (datetime): Time now
            last_timestamp (datetime): Timestamp of the last output
            timerange_value (int): Value of the timerange in the configuration
            timerange_type (str): Type of the timerange (absolute or relative)

        Returns:
            tuple[str, str]: Timestamps for the input data query (from_date, to_date)
        """
        if timerange_type is TimerangeTypes.ABSOLUTE or last_timestamp is None:
            from_date = (time_now - timedelta(seconds=timerange_value)).strftime(
                "%Y-%m-%dT%H:%M:%S%z"
            )
            return from_date, None

        timeframe = (time_now - last_timestamp).total_seconds() / 60

        if timerange_type is TimerangeTypes.RELATIVE:
            if timeframe < timerange_value:
                from_date = (time_now - timedelta(seconds=timerange_value)).strftime(
                    "%Y-%m-%dT%H:%M:%S%z"
                )
                return from_date, None

            from_date = last_timestamp.strftime("%Y-%m-%dT%H:%M:%S%z")
            to_date = (last_timestamp + timedelta(seconds=timerange_value)).strftime(
                "%Y-%m-%dT%H:%M:%S%z"
            )
            return from_date, to_date

        # Fallback to absolute if no type is specified
        from_date = (time_now - timedelta(seconds=timerange_value)).strftime(
            "%Y-%m-%dT%H:%M:%S%z"
        )
        return from_date, None

    def _calculate_timerange_min_max(
        self,
        time_now: datetime,
        last_timestamp: Union[datetime, None],
        timerange_min: int,
        timerange_max: int,
    ) -> tuple[str, str]:
        """Function to calculate the timerange for the input data query based on a min
        and max timerange from the configuration

        Args:
            time_now (datetime): Time now
            last_timestamp (datetime): Timestamp of the last output, if available
            timerange_min (int): Minimal value of the timerange in the configuration in seconds
            timerange_max (int): Maximal value of the timerange in the configuration in seconds
            timerange_type (TimerangeTypes): Type of the timerange (absolute or relative)

        Returns:
            tuple[str, str]: Timestamps for the input data query (from_date, to_date)
        """
        if last_timestamp is None:
            from_date = (
                (time_now - timedelta(seconds=timerange_max))
                .replace(tzinfo=tz.UTC)
                .strftime("%Y-%m-%dT%H:%M:%S%z")
            )
            return from_date, None

        timeframe = (time_now - last_timestamp).total_seconds() / 60

        if timeframe < timerange_min:
            from_date = (
                (time_now - timedelta(seconds=timerange_min))
                .replace(tzinfo=tz.UTC)
                .strftime("%Y-%m-%dT%H:%M:%S%z")
            )
            return from_date, None

        if timeframe < timerange_max:
            from_date = last_timestamp.strftime("%Y-%m-%dT%H:%M:%S%z")
            return from_date, None

        from_date = last_timestamp.strftime("%Y-%m-%dT%H:%M:%S%z")
        to_date = (last_timestamp + timedelta(seconds=timerange_max)).strftime(
            "%Y-%m-%dT%H:%M:%S%z"
        )
        return from_date, to_date

    def _handle_calculation_method(
        self,
        time_now: datetime,
        last_timestamp: datetime,
    ) -> tuple[str, str]:
        """Funtion to calculate the dates for the calculation method

        Args:
            time_now (datetime): Time now
            last_timestamp (datetime): Timestamp of the last output

        Returns:
            tuple[str, str]: Timestamps for the input data query (from_date, to_date)
        """
        calculation = self.config.controller_settings.time_settings.calculation

        if calculation.timerange is not None:
            return self._calculate_timerange(
                time_now,
                last_timestamp,
                calculation.timerange
                * get_time_unit_seconds(calculation.timerange_unit),
                calculation.timerange_type,
            )
        if (
            calculation.timerange_min is not None
            and calculation.timerange_max is not None
        ):
            if calculation.timerange_type is TimerangeTypes.ABSOLUTE:
                return self._calculate_timerange(
                    time_now,
                    last_timestamp,
                    calculation.timerange_max
                    * get_time_unit_seconds(calculation.timerange_unit),
                    calculation.timerange_type,
                )

            return self._calculate_timerange_min_max(
                time_now,
                last_timestamp,
                calculation.timerange_min
                * get_time_unit_seconds(calculation.timerange_unit),
                calculation.timerange_max
                * get_time_unit_seconds(calculation.timerange_unit),
            )

        logger.error(
            "No Information about the input time ranges for the calculation in the configuration."
        )
        return None, None

    def _calculate_dates(
        self, method: DataQueryTypes, last_timestamp: Union[datetime, None]
    ) -> tuple[str, str]:
        """Function to calculate the dates for the input data query

        Args:
            method (DataQueryTypes): Method for the calculation
            time_now (datetime): Time now
            last_timestamp (datetime): Timestamp of the last output

        Returns:
            tuple[str, str]: Timestamps for the input data query (from_date, to_date)
        """

        time_now = datetime.now(timezone.utc)

        from_date, to_date = None, None

        if method is DataQueryTypes.CALCULATION:
            from_date, to_date = self._handle_calculation_method(
                time_now, last_timestamp
            )
        elif method is DataQueryTypes.CALIBRATION:
            from_date, to_date = self._handle_calibration_method(
                time_now, last_timestamp
            )

        if to_date is None:
            to_date = time_now.strftime("%Y-%m-%dT%H:%M:%S%z")

        return from_date, to_date

    def _handle_calibration_method(
        self, time_now: datetime, last_timestamp: Union[datetime, None]
    ) -> tuple[str, str]:
        """Funtion to calculate the dates for the calibration method
            - Use the last timestamp of the output entity as the from_date if available
            and time range is relative
            - Use the timenow - timerange as the from_date if the time range is absolute
            or no last timestamp is available

        Args:
            time_now (datetime): Time now
            last_timestamp (datetime): Timestamp of the last output, if available

        Returns:
            tuple[str, str]: Timestamps for the input data query (from_date, to_date)
        """
        calibration_time_settings = (
            self.config.controller_settings.time_settings.calibration
        )

        timerange = calibration_time_settings.timerange * get_time_unit_seconds(
            calibration_time_settings.timerange_unit
        )

        if (
            self.config.controller_settings.time_settings.calculation.timerange_type
            is TimerangeTypes.RELATIVE
            and last_timestamp is not None
        ):
            from_date = (last_timestamp - timedelta(seconds=timerange)).strftime(
                "%Y-%m-%dT%H:%M:%S%z"
            )
            to_date = last_timestamp

            return from_date, None

        from_date = (time_now - timedelta(seconds=timerange)).strftime(
            "%Y-%m-%dT%H:%M:%S%z"
        )
        to_date = time_now

        return from_date, to_date

    def get_data_from_datebase(
        self,
        entity: ContextEntity,
        entity_attributes: dict,
        method: DataQueryTypes,
        timestamp_latest_output: datetime,
    ) -> list[InputDataAttributeModel]:
        """
        Function to get the data from the database for the input attributes (crateDB is used)

        Args:
            - entity_id: id of the entity
            - entity_type: type of the entity
            - entity_attributes: dict with the attributes of the entity (id, metadata)
            - method (DataQueryTypes): method of the function which queries the data
            (calculation or calibration)
            - timestamp_latest_output: timestamp of the last output of the entity

        Returns:
            - list of InputDataAttributeModel: list with the input attributes

        TODO:
            - Does it make sense to use the quantumleap client for this?
                https://github.com/RWTH-EBC/FiLiP/blob/master/filip/clients/ngsi_v2/quantumleap.py#L449
            - Improve the error handling
        """

        from_date, to_date = self._calculate_dates(
            method=method, last_timestamp=timestamp_latest_output
        )

        df = self.crate_db_client.get_data(
            service=self.fiware_conn_params.fiware_params.service,
            entity=entity,
            attributes=[
                attribute["id_interface"] for attribute in entity_attributes.values()
            ],
            from_date=from_date,
            to_date=to_date,
            limit=1000000,
        )

        if df.empty:
            logger.debug("Service has not received data from CrateDB")
            return []

        # resample the time series with configured time step size
        df.fillna(value=np.nan, inplace=True)
        time_step_seconds = int(
            self.config.controller_settings.time_settings.calculation.timestep
            * get_time_unit_seconds(
                self.config.controller_settings.time_settings.calculation.timestep_unit
            )
        )
        df = df.resample(f"""{time_step_seconds}s""").mean(numeric_only=True)

        input_attributes = []

        for attribute_id, attribute_data in entity_attributes.items():
            df.rename(
                columns={attribute_data["id_interface"]: attribute_id}, inplace=True
            )
            data = df.filter([attribute_id]).dropna()
            if data.empty:
                logger.debug(
                    f"Data for attribute {attribute_id} of entity {entity.id} is empty"
                )
                input_attributes.append(
                    InputDataAttributeModel(
                        id=attribute_id,
                        data=None,
                        data_type=AttributeTypes.TIMESERIES,
                        data_available=False,
                        latest_timestamp_input=None,
                        unit=attribute_data["metadata"].unit,
                    )
                )
            else:
                logger.debug(
                    f"Service received data from CrateDB for attribute {attribute_id} "
                    f"of entity {entity.id}"
                )
                input_attributes.append(
                    InputDataAttributeModel(
                        id=attribute_id,
                        data=data,
                        data_type=AttributeTypes.TIMESERIES,
                        data_available=True,
                        latest_timestamp_input=to_date,
                        unit=attribute_data["metadata"].unit,
                    )
                )

        return input_attributes

    def update_fiware_entity(
        self,
        entity_id: str,
        entity_type: str,
        attrs: list[NamedContextAttribute],
    ) -> None:
        """
        Function to update the entity in the FIWARE platform with filip

        Args:
            entity_id (str): ID of the entity
            entity_type (str): Type of the entity
            attrs (list[NamedContextAttribute]): List with the attributes to update
        """

        self.cb_client.update_or_append_entity_attributes(
            entity_id=entity_id, entity_type=entity_type, attrs=attrs
        )

    async def _send_timeseries_to_fiware(
        self,
        entity_id: str,
        entity_type: str,
        attrs_timeseries: list[NamedContextAttribute],
    ) -> None:
        """
        Function to send the timeseries data to the FIWARE platform in async mode
        and parallel processing

        Args:
            entity_id (str): ID of the entity
            entity_type (str): Type of the entity
            attrs_timeseries (list[NamedContextAttribute]): List with the timeseries data
        TODO:
            - Is there a better way to send the data from dataframes to the FIWARE platform?
        """
        max_workers = multiprocessing.cpu_count()

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for attribute in attrs_timeseries:
                future = executor.submit(
                    self.update_fiware_entity, entity_id, entity_type, [attribute]
                )
                futures.append(future)

                await sleep(0.01)

            concurrent.futures.wait(futures)

    async def prepare_timeseries_for_fiware(
        self,
        fiware_datapoint: FiwareDatapointParameter,
        datatype: DataType,
        factor_unit_adjustment: float,
    ) -> list[NamedContextAttribute]:
        """
        Function to prepare the timeseries data for the FIWARE platform

        Args:
            fiware_datapoint (FiwareDatapointParameter): Fiware datapoint parameter
            datatype (DataType): Datatype of the attribute
            factor_unit_adjustment (float): Factor to adjust the unit

        Returns:
            list: List with the attributes (NamedContextAttribute) for the FIWARE platform
        """
        if len(fiware_datapoint.attribute.value) == 0:
            return
        if (
            fiware_datapoint.attribute.id
            not in fiware_datapoint.attribute.value.columns
        ):
            logger.error(
                f"Attribute {fiware_datapoint.attribute.id} not in the dataframe."
            )
            return
        df = fiware_datapoint.attribute.value.sort_index()
        attrs_timeseries = []

        meta_data_row = fiware_datapoint.metadata + [
            NamedMetadata(
                name="TimeInstant",
                type=DataType.DATETIME,
                value=df.index[-1].strftime("%Y-%m-%dT%H:%M:%S%z"),
            )
        ]

        attr = NamedContextAttribute(
            name=fiware_datapoint.attribute.id_interface,
            value=df[fiware_datapoint.attribute.id].iloc[-1] * factor_unit_adjustment,
            type=datatype,
            metadata=meta_data_row,
        )

        for index, row in fiware_datapoint.attribute.value.iterrows():
            if index == df.index[-1]:
                continue

            meta_data_row = fiware_datapoint.metadata + [
                NamedMetadata(
                    name="TimeInstant",
                    type=DataType.DATETIME,
                    value=index.strftime("%Y-%m-%dT%H:%M:%S%z"),
                )
            ]

            attrs_timeseries.append(
                NamedContextAttribute(
                    name=fiware_datapoint.attribute.id_interface,
                    value=row[fiware_datapoint.attribute.id] * factor_unit_adjustment,
                    type=datatype,
                    metadata=meta_data_row,
                )
            )
        if len(attrs_timeseries) > 0:
            await self._send_timeseries_to_fiware(
                entity_id=fiware_datapoint.entity.id,
                entity_type=fiware_datapoint.entity.type,
                attrs_timeseries=attrs_timeseries,
            )

        return attr

    async def _send_data_to_fiware(
        self,
        output_entity: OutputModel,
        output_attributes: list[AttributeModel],
        output_commands: list[CommandModel],
    ) -> None:
        """
        Function to send the output data to the FIWARE platform

        Args:
            - output_entity: OutputModel with the output entity
            - output_attributes: list with the output attributes
            - output_commands: list with the output commands

        TODO:
            - Is there a better way to send the data from dataframes to the FIWARE platform?
        """

        fiware_entity = self.cb_client.get_entity(output_entity.id_interface)

        entity_attributes = self.cb_client.get_entity_attributes(
            entity_id=fiware_entity.id, entity_type=fiware_entity.type
        )

        attrs = []
        for attribute in output_attributes:

            fiware_unit = None
            factor_unit_adjustment = 1

            if attribute.id_interface in entity_attributes:
                datatype = entity_attributes[attribute.id_interface].type
                if (
                    entity_attributes[attribute.id_interface].metadata.get("unitCode")
                    is not None
                ):
                    fiware_unit = DataUnits(
                        entity_attributes[attribute.id_interface]
                        .metadata.get("unitCode")
                        .value
                    )
            else:
                datatype = attribute.datatype

            meta_data = []

            if attribute.unit is not None and fiware_unit is None:
                meta_data.append(
                    NamedMetadata(
                        name="unitCode", type=DataType.TEXT, value=attribute.unit.value
                    )
                )
            elif attribute.unit is None:
                logger.debug(
                    f"No information about the unit of the attribute {attribute.id} "
                    f"from entity {output_entity.id} available!"
                )

            elif fiware_unit is not attribute.unit:

                factor_unit_adjustment = get_unit_adjustment_factor(
                    unit_actual=attribute.unit, unit_target=fiware_unit
                )
            if isinstance(attribute.value, pd.DataFrame):

                attrs.append(
                    await self.prepare_timeseries_for_fiware(
                        fiware_datapoint=FiwareDatapointParameter(
                            entity=ContextEntity(
                                id=fiware_entity.id, type=fiware_entity.type
                            ),
                            attribute=attribute,
                            metadata=meta_data,
                        ),
                        datatype=datatype,
                        factor_unit_adjustment=factor_unit_adjustment,
                    )
                )
                continue
            meta_data.append(
                NamedMetadata(
                    name="TimeInstant",
                    type=DataType.DATETIME,
                    value=attribute.timestamp.strftime("%Y-%m-%dT%H:%M:%S%z"),
                )
            )
            attrs.append(
                NamedContextAttribute(
                    name=attribute.id_interface,
                    value=(
                        attribute.value * factor_unit_adjustment
                        if attribute.value is not None
                        else None
                    ),
                    type=datatype,
                    metadata=meta_data,
                )
            )

        cmds = []
        for command in output_commands:
            cmds.append(
                NamedCommand(
                    name=command.id_interface,
                    value=command.value,
                    type=DataType.COMMAND,
                )
            )

        output_points = attrs + cmds

        if len(output_points) == 0:
            logger.debug(
                "There is no output data available to send to the FIWARE platform "
                f"for the entity {fiware_entity.id}."
            )
            return

        if len(attrs) > 0:
            i = 0
            while i < 3:
                try:
                    self.cb_client.update_or_append_entity_attributes(
                        entity_id=fiware_entity.id,
                        entity_type=fiware_entity.type,
                        attrs=attrs,
                    )
                    break
                except requests.exceptions.HTTPError as err:
                    if i < 3:
                        await sleep(0.1)
                    else:
                        logger.error(
                            f"HTTPError while sending attributes to FIWARE platform: {err}"
                        )

                i += 1

        if len(cmds) > 0:
            i = 0
            while i < 3:
                try:
                    self.cb_client.update_existing_entity_attributes(
                        entity_id=fiware_entity.id,
                        entity_type=fiware_entity.type,
                        attrs=cmds,
                    )
                    break
                except requests.exceptions.HTTPError as err:
                    if i < 3:
                        await sleep(0.1)
                    else:
                        logger.error(
                            f"HTTPError while sending commands to FIWARE platform: {err}"
                        )

                i += 1



"""
Description: This file contains the units for the use and conversion
of different units and in the system controller.
Author: Martin Altenburger
"""

import datetime
from enum import Enum
from typing import Union, Optional

from loguru import logger


class TimeUnits(Enum):
    """Possible time units for the time series data

    TODO: Is it better to use standard time units? Like in the unit code?

    """

    SECOND = "second"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    MONTH = "month"


class TimeUnitsSeconds(Enum):
    """Seconds for the time units"""

    SECOND = datetime.timedelta(seconds=1).total_seconds()
    MINUTE = datetime.timedelta(minutes=1).total_seconds()
    HOUR = datetime.timedelta(hours=1).total_seconds()
    DAY = datetime.timedelta(days=1).total_seconds()
    MONTH = datetime.timedelta(days=30).total_seconds()


class DataUnits(Enum):
    """
    Possible units for the data
    Units which are defined by Unit Code (https://unece.org/trade/cefact/UNLOCODE-Download
    or https://github.com/RWTH-EBC/FiLiP/blob/master/filip/data/unece-units/units_of_measure.csv)
    or here: https://unece.org/fileadmin/DAM/cefact/recommendations/rec20/rec20_rev3_Annex3e.pdf
    TODO:
        - Is there a better way to handle the units?
        - Add more units?

    """

    # Time
    SECOND = "SEC"  # "seconds"
    HOUR = "HUR"  # "hour"
    MINUTE = "MIN"  # "minute"

    # Temperature
    DEGREECELSIUS = "CEL"  # "°C"
    KELVIN = "KEL"  # "K"

    # Volume / Volumeflow
    LITER = "LTR"  # "l"
    MTQ = "MTQ"  # "m³"
    MQH = "MQH"  # "m³/h"
    MQS = "MQS"  # "m³/s"
    E32 = "E32"  # "l/h" (E32 is the unit code for liters per hour)
    L2 = "L2"  # "l/m" (L2 is the unit code for liters per minute)

    # Energy / Power
    WTT = "WTT"  # "W"
    WHR = "WHR"  # "Wh"
    KWH = "KWH"  # "kWh"

    # Distance/ Area
    CMT = "CMT"  # "cm"
    MTR = "MTR"  # "m"
    MTK = "MTK"  # "m²"

    # speed
    MTS = "MTS"  # "m/s"

    # unitless
    PERCENT = "P1"  # "%"

    # Electrical
    OHM = "OHM"  # "Ohm"
    VLT = "VLT"  # "V"


def get_unit_adjustment_factor(
    unit_actual: DataUnits, unit_target: DataUnits
) -> Optional[float]:
    """Function to get the adjustment factor for the conversion of units

    Args:
        unit_actual (DataUnits): Actual unit
        unit_target (DataUnits): Target unit

    Returns:
        Optional[float]: Adjustment factor for the conversion of the units, if found
    """

    if unit_actual is None:
        logger.warning("Actual unit is None, could not determine adjustment factor")
        return None
    if unit_target is None:
        logger.warning("Target unit is None, could not determine adjustment factor")
        return None
    if unit_actual == unit_target:
        return 1.0
    # TODO: Real adjustment factors
    logger.warning(
        f"Adjustment factor for the conversion of {unit_actual} to {unit_target} "
        "not implemented yet"
    )
    return None


def get_time_unit_seconds(
    time_unit: Union[TimeUnits, str, DataUnits],
) -> Optional[float]:
    """Funktion to get the seconds for a time unit

    Args:
        time_unit (Union[TimeUnits, str, DataUnits]): time unit / Name of the time unit \
            If you use DataUnits, only the units which are also in TimeUnits are valid

    Returns:
        Union[int, None]: Number of seconds for the time unit\
            or None if the time unit is not available
    """
    if isinstance(time_unit, TimeUnits):
        return TimeUnitsSeconds[time_unit.name].value

    if time_unit in [unit.value for unit in TimeUnits]:
        return TimeUnitsSeconds[TimeUnits(time_unit).name].value

    if isinstance(time_unit, DataUnits):
        return TimeUnitsSeconds[TimeUnits[time_unit.name].name].value

    if time_unit in [unit.name for unit in DataUnits]:
        if DataUnits(time_unit).name in [unit.name for unit in TimeUnits]:
            return TimeUnitsSeconds[TimeUnits(DataUnits(time_unit).name).name].value

    logger.warning(f"Time unit {time_unit} not available")
    return None



"""
Description: Default values for environment variables of the service
Authors: Martin Altenburger
"""

from enum import Enum


class DefaultEnvVariables(Enum):
    """
    Default environment variables for service
    """

    CONFIG_PATH = "./config.json"
    CB_URL = "http://localhost:1026"
    FIWARE_SERVICE = "service"
    FIWARE_SERVICE_PATH = "/"
    FIWARE_AUTH = False
    CRATE_DB_URL = "http://localhost:4200"
    CRATE_DB_USER = "crate"
    CRATE_DB_PW = ""
    CRATE_DB_SSL = False
    LOG_LEVEL = "WARNING"

    PATH_OF_INPUT_FILE = "./config/Input_file.csv"
    PATH_OF_STATIC_DATA = "../static_data.json"
    START_TIME_FILE = "01.01.2023 00:00"  # TODO: IsoFormat is better?
    TIME_FORMAT_FILE = "%d.%m.%Y %H:%M"
    RELOAD_STATICDATA = False

    MQTT_HOST = "localhost"
    MQTT_PORT = 1883
    MQTT_USERNAME = ""
    MQTT_PASSWORD = ""
    MQTT_TOPIC_PREFIX = ""  # used as prefix for all topics



"""
Basic configuration for the components in the EnCoCaPy framework.
Author: Martin Altenburger
"""

from typing import Dict, Optional

from loguru import logger
from pydantic import BaseModel, Field, RootModel, model_validator

from encodapy.utils.units import DataUnits, get_unit_adjustment_factor
from encodapy.utils.datapoints import DataPointGeneral


# Models for the Input Configuration
class IOAllocationModel(BaseModel):
    """
    Model for the input or output allocation.

    Contains:
        `entity`: ID of the entity to which the input or output is allocated
        `attribute`: ID of the attribute to which the input or output is allocated
    """

    entity: str = Field(
        ..., description="ID of the entity to which the input or output is allocated"
    )
    attribute: str = Field(
        ..., description="ID of the attribute to which the input or output is allocated"
    )


class IOModell(
    (RootModel[Dict[str, IOAllocationModel]])
):  # pylint: disable=too-few-public-methods
    """
    Model for the input, staticdata and output of a component.

    It contains a dictionary with the key as the ID of the input, output or static data
    and the value as the allocation model.

    There is no validation for this.
    It is used to create the the ComponentIOModel for each component.
    """


class ConfigDataPoints(
    (RootModel[Dict[str, IOAllocationModel | DataPointGeneral]])
):  # pylint: disable=too-few-public-methods
    """
    Model for the configuration of config data points.
    """


class ControllerComponentModel(BaseModel):
    """
    Model for the configuration of the controller components.
    Contains:
    - active: Whether the component is active or not
    - id: The id of the component
    - type: The type of the component (e.g. thermal storage, heat pump, etc. / \
        needs to be defined for individual components)
    - inputs: The inputs of the component as a dictionary with IOAllocationModel \
        for the individual inputs
    - outputs: The outputs of the component as a dictionary with IOAllocationModel \
        for the individual outputs
    - config: The configuration of the component as a dictionary with IOAllocationModel \
          for the individual static data or DataPointModel with direct values
    """

    active: Optional[bool] = True
    id: str
    type: str
    inputs: IOModell
    outputs: IOModell
    config: Optional[ConfigDataPoints] = None


# Models for the internal input and output connections, needs to filled for the components
class ComponentData(BaseModel):
    """
    Basemodel for the configuration of the datapoints of a component

    Base for InputData, OutputData and ConfigData
    
    Provides a validator to check the units of the input values \
        and convert them if necessary.
    """

    @model_validator(mode="after")
    def check_unit_values(self) -> "ComponentData":
        """
        Check the units of the input values and convert them if necessary.
        """
        for name, field in self.model_fields.items():
            value = getattr(self, name)
            extra = field.json_schema_extra or {}
            if not isinstance(extra, dict):
                logger.warning(f"Extra for field {name} is not a dictionary: {extra}")
                continue

            if isinstance(value, DataPointGeneral):
                if "unit" in extra.keys() and value.unit is None:
                    value.unit = DataUnits(extra["unit"])
                elif (
                    "unit" in extra.keys()
                    and value.unit is not None
                    and value.unit != DataUnits(extra["unit"])
                ):
                    if value.value is None or not isinstance(value.value, (int, float)):
                        logger.warning(
                            f"Unit of {name} is {value.unit}, but expected {extra['unit']}. "
                            f"Could not convert, because value is None or not a number."
                        )
                        continue
                    unit_adjustment_factor = get_unit_adjustment_factor(
                        unit_actual=value.unit, unit_target=DataUnits(extra["unit"])
                    )
                    if unit_adjustment_factor is None:
                        logger.warning(
                            f"Unit of {name} is {value.unit}, but expected {extra['unit']}. "
                            f"Could not convert, because units are not compatible "
                            "or no adjustment factor found."
                        )
                        continue
                    value.value = value.value * unit_adjustment_factor
                    value.unit = DataUnits(extra["unit"])
        return self


class OutputData(ComponentData):
    """
    Basemodel for the configuration of the outputs of a component

    Needs to be implemented by the user.
    
    Provides a validator to check the units of the input values \
        and convert them if necessary.
    """


class InputData(ComponentData):
    """
    Basemodel for the configuration of the inputs of a component
    
    Needs to be implemented by the user.
    
    Provides a validator to check the units of the input values \
        and convert them if necessary.
    Only works if the value is a number (int or float) and the datapoint \
        is a DataPointGeneral or a SubModel.
    """


class ConfigData(ComponentData):
    """
    Basemodel for the configuration data of a component
    
    Needs to be implemented by the user, if static configuration is needed.
    
    Provides a validator to check the units of the input values \
        and convert them if necessary.
    """


class ComponentIOModel(BaseModel):
    """
    Model for the input and output of the thermal storage service.

    Contains:
        `input`: InputModel = Input configuration for the thermal storage service
        `output`: OutputModel = Output configuration for the thermal storage service
    """

    input: InputData = Field(
        ..., description="Input configuration for the thermal storage service"
    )
    output: OutputData = Field(
        ..., description="Output configuration for the thermal storage service"
    )


# Custom Exceptions
class ComponentValidationError(Exception):
    """Custom error for invalid configurations."""



"""
Description: Simple component of a two-point controller
Author: Martin Altenburger
"""

from typing import Optional, Union
from loguru import logger
from encodapy.components.basic_component import BasicComponent
from encodapy.components.basic_component_config import ControllerComponentModel
from encodapy.components.two_point_controller.two_point_controller_config import (
    TwoPointControllerConfigData,
    TwoPointControllerInputData,
    TwoPointControllerOutputData,
)
from encodapy.utils.models import (
    StaticDataEntityModel,
)
from encodapy.utils.datapoints import DataPointGeneral


class TwoPointController(BasicComponent):
    """
    Class for a two-point controller
    Args:
        config (Union[ControllerComponentModel, list[ControllerComponentModel]]): \
            The configuration for the controller.
        component_id (str): The unique identifier for the component.
        static_data (Optional[list[StaticDataEntityModel]]): Static data for the component.
    """

    def __init__(
        self,
        config: Union[ControllerComponentModel, list[ControllerComponentModel]],
        component_id: str,
        static_data: Optional[list[StaticDataEntityModel]] = None,
    ):
        self.config_data: TwoPointControllerConfigData
        self.input_data: TwoPointControllerInputData
        self.output_data: TwoPointControllerOutputData

        super().__init__(
            component_id=component_id, config=config, static_data=static_data
        )

    def get_control_signal(
        self,
    ) -> DataPointGeneral:
        """Calculate the control signal based on current and setpoint values."""

        minimal_value = (
            self.config_data.setpoint.value - self.config_data.hysteresis.value
        )

        # Raise errors if units do not match
        try:
            assert self.config_data.hysteresis.unit == self.config_data.setpoint.unit, (
                f"Units of hysteresis ({self.config_data.hysteresis.unit}) "
                f"and setpoint ({self.config_data.setpoint.unit}) must be the same!"
            )
            assert (
                self.input_data.current_value.unit == self.config_data.setpoint.unit
            ), (
                f"Units of current_value ({self.input_data.current_value.unit}) "
                f"and setpoint ({self.config_data.setpoint.unit}) must be the same!"
            )
            assert (
                self.input_data.latest_control_signal.unit
                == self.config_data.command_enabled.unit
            ), (
                f"Units of latest_control_signal ({self.input_data.latest_control_signal.unit}) "
                f"and command_enabled ({self.config_data.command_enabled.unit}) must be the same!"
            )
            assert (
                self.input_data.latest_control_signal.unit
                == self.config_data.command_disabled.unit
            ), (
                f"Units of latest_control_signal ({self.input_data.latest_control_signal.unit}) "
                f"and command_disabled ({self.config_data.command_disabled.unit}) must be the same!"
            )
        except AssertionError as e:
            logger.error(f"Unit assertion error in {self.component_config.id}: {e}")
            raise

        if self.input_data.current_value.value < minimal_value:
            return DataPointGeneral(
                value=self.config_data.command_enabled.value,
                unit=self.config_data.command_enabled.unit,
            )

        if self.input_data.current_value.value > float(self.config_data.setpoint.value):
            return DataPointGeneral(
                value=self.config_data.command_disabled.value,
                unit=self.config_data.command_disabled.unit,
            )

        if (
            self.input_data.latest_control_signal.value
            == self.config_data.command_enabled.value
            and self.input_data.current_value.value > minimal_value
        ):
            return DataPointGeneral(
                value=self.config_data.command_enabled.value,
                unit=self.config_data.command_enabled.unit,
            )

        return DataPointGeneral(
            value=self.config_data.command_disabled.value,
            unit=self.config_data.command_disabled.unit,
        )

    def calculate(self):
        """
        Calculate the output data based on the input data and configuration.
        """

        self.output_data = TwoPointControllerOutputData(
            control_signal=self.get_control_signal()
        )



"""
Description: This module provides basic components for the encodapy package.
Author: Martin Altenburger
"""

from datetime import datetime, timezone
from typing import Any, Optional, Type, Union, TypeVar, Generic, cast
from loguru import logger
from pydantic import ValidationError
from encodapy.components.basic_component_config import (
    ComponentIOModel,
    ComponentValidationError,
    ConfigData,
    ConfigDataPoints,
    ControllerComponentModel,
    DataPointGeneral,
    InputData,
    IOAllocationModel,
    IOModell,
    OutputData,
)
from encodapy.components.component_loader import (
    get_component_config_data_model,
    get_component_input_data_model,
    get_component_io_model,
    get_component_output_data_model,
)
from encodapy.utils.models import (
    DataTransferComponentModel,
    InputDataEntityModel,
    InputDataModel,
    StaticDataEntityModel,
)

# Type variables for component data models
# - these are used for type hinting in the BasicComponent class
TypeConfigData = TypeVar(
    "TypeConfigData", bound=ConfigData
)  # pylint: disable=invalid-name
TypeInputData = TypeVar(
    "TypeInputData", bound=InputData
)  # pylint: disable=invalid-name
TypeOutputData = TypeVar(
    "TypeOutputData", bound=OutputData
)  # pylint: disable=invalid-name


class BasicComponent(Generic[TypeConfigData, TypeInputData, TypeOutputData]):
    """
    Base class for all components in the encodapy package.
    This class provides basic functionality that can be extended by specific components.

    Contains methods for:
    - Getting component configuration: `get_component_config()`
    - Getting input values: `get_component_input()`
    - Setting all static data of the component: `set_component_static_data()`
    - Getting static data (by id): `get_component_static_data()`

    Args:
        config (Union[ControllerComponentModel, list[ControllerComponentModel]]):
            Configuration of the component or a list of configurations.
        component_id (str): ID of the component to get the configuration for.
        static_data (Optional[list[StaticDataEntityModel]]): Static data for the component.
    """

    def __init__(
        self,
        config: Union[ControllerComponentModel, list[ControllerComponentModel]],
        component_id: str,
        static_data: Optional[list[StaticDataEntityModel]] = None,
    ) -> None:
        if isinstance(config, ControllerComponentModel):
            self.component_config = ControllerComponentModel.model_validate(config)
        else:
            self.component_config = ControllerComponentModel.model_validate(
                self.get_component_config(config=config, component_id=component_id)
            )

        self.config_data: TypeConfigData
        self.set_component_config_data(
            static_data=static_data, static_config=self.component_config.config
        )
        # Inputs and Outputs of the component itself
        self.io_model: Optional[ComponentIOModel] = None
        self.input_data: TypeInputData
        self.output_data: TypeOutputData

        self._prepare_i_o_config()

        self.prepare_component()

    def get_component_config(
        self, config: list[ControllerComponentModel], component_id: str
    ) -> ControllerComponentModel:
        """
        Function to get the configuration of a specific component from the service configuration
        Args:
            config (list[ControllerComponentModel]): List of all components in the configuration
            component_id (str): ID of the component to get the configuration
        Returns:
            ControllerComponentModel: Configuration of the component by ID

        Raises:
            ValueError: If the component with the given ID is not found in the configuration
        """
        for component in config:
            if component.id == component_id:
                return component

        raise KeyError(f"No component configuration found for {component_id}")

    def _get_input_and_output_config_models(
        self,
    ) -> tuple[Type[InputData], Type[OutputData]]:
        """
        Function to get the input and output models for the component.
        There needs to be a InputModel and a OutputModel in the config-module for the component.

        """
        component_input_model = get_component_io_model(
            component_type=self.component_config.type, model_subname="InputData"
        )
        component_output_model = get_component_io_model(
            component_type=self.component_config.type, model_subname="OutputData"
        )

        if not (
            issubclass(component_input_model, InputData)
            and issubclass(component_output_model, OutputData)
        ):
            error_msg = "Input or output model is not a subclass of BaseModel"
            logger.error(error_msg)
            raise TypeError(error_msg)

        return component_input_model, component_output_model

    def _prepare_i_o_config(self):
        """
        Function to prepare the I/O configuration for the component
        """
        component_input_model, component_output_model = (
            self._get_input_and_output_config_models()
        )
        config = self.component_config
        try:
            input_config = component_input_model.model_validate(
                config.inputs.root
                if isinstance(config.inputs, IOModell)
                else config.inputs
            )
        except ValidationError:
            error_msg = f"Invalid input configuration for the component {self.component_config.id}"
            logger.error(error_msg)
            raise

        try:
            output_config = component_output_model.model_validate(
                config.outputs.root
                if isinstance(config.outputs, IOModell)
                else config.outputs
            )
        except ValidationError:
            error_msg = f"Invalid output configuration for the component {self.component_config.id}"
            logger.error(error_msg)
            raise

        self.io_model = ComponentIOModel(input=input_config, output=output_config)

    def set_component_config_data(
        self,
        static_data: Union[list[StaticDataEntityModel], None],
        static_config: Optional[ConfigDataPoints] = None,
    ):
        """
        Function to get the value of the static data for a specific input configuration \
            of a component of the controller (or a individual one).

        Args:
            config_data (Union[list[StaticDataEntityModel], None]): Data of static entities
            static_config (Optional[ConfigDataPoints]): \
                Configuration of the static data, if available
                
        Raises:
            ComponentValidationError: If the static data configuration is invalid.

        """
        try:
            assert (
                static_config is not None
            ), "No static config provided, skipping static data setup."
            assert isinstance(
                static_config, ConfigDataPoints
            ), "Invalid static config provided."
        except AssertionError as e:
            logger.error(f"Static config error: {e}")
            raise ComponentValidationError(f"Static config error: {e}") from e

        config_model = get_component_config_data_model(
            component_type=self.component_config.type, model_subname="ConfigData"
        )

        if config_model is None:
            logger.debug("No config model found, skipping static config check.")
            return

        static_config_data: dict[str, DataPointGeneral] = {}

        for datapoint_name, _ in config_model.model_fields.items():
            if datapoint_name not in static_config.root:
                logger.debug(
                    f"Static config {datapoint_name} not in static configuration, "
                    "trying to skip it."
                )
                continue
            datapoint = static_config.root[datapoint_name]

            if isinstance(datapoint, DataPointGeneral):
                static_config_data[datapoint_name] = datapoint

            if isinstance(datapoint, IOAllocationModel):
                if static_data is None:
                    error_msg = (
                        f"Config entry '{datapoint_name}' needs static data but its not provided "
                        f"to the component {self.component_config.id}."
                    )
                    logger.error(error_msg)
                    raise ComponentValidationError(error_msg)
                static_config_data[datapoint_name] = self.get_component_input(
                    input_entities=static_data, input_config=datapoint
                )

        # we need to convert the data to a dict of the correct types
        # because the config data model could contain different types
        static_config_raw: dict[str, Any] = {}
        for key, value in static_config_data.items():
            static_config_raw[key] = value.model_dump()

        try:
            self.config_data = cast(
                TypeConfigData, config_model.model_validate(static_config_raw)
            )

        except ValidationError as error:
            error_msg = (
                f"Error in static data configuration: {error}"
                " Could not validate and set the static data model"
            )
            logger.error(error_msg)
            raise ComponentValidationError(error_msg) from error

    def get_component_input(
        self,
        input_entities: Union[
            list[InputDataEntityModel],
            list[StaticDataEntityModel],
            list[Union[InputDataEntityModel, StaticDataEntityModel]],
        ],
        input_config: IOAllocationModel,
    ) -> DataPointGeneral:
        """
        Function to get the value of the input data for a specific input configuration \
            of a component of the controller (or a individual one).

        Args:
            input_entities (list[InputDataEntityModel]): Data of input entities
            input_config (IOAllocationModel): Configuration of the input

        Returns:
            DataPointGeneral: The value of the input data and its unit and timestamp
        """
        for input_data in input_entities:
            if input_data.id == input_config.entity:
                for attribute in input_data.attributes:
                    if attribute.id == input_config.attribute:
                        return DataPointGeneral(
                            value=attribute.data,
                            unit=attribute.unit,
                            time=attribute.latest_timestamp_input,
                        )

        raise KeyError(
            f"Input data {input_config.entity} / {input_config.attribute} not found. "
            "Please check the configuration of the Inputs, Outputs and Static Data."
        )

    def set_input_data(self, input_data: InputDataModel) -> None:
        """
        Set the input values for the component from the provided input entities.
        The input values are extracted based on the component's input configuration.
        Also static data is used as input.

        Input data is validated against the component's input model and stored in self.input_data.

        The validation ensures that the input data conforms to the expected structure and types.
        Also the units are checked and converted if necessary.

        Args:
            input_data (InputDataModel): Input data model containing all necessary entities

        """
        if self.io_model is None:
            return
        # use all input datapoints, also the static data (flexible solution)
        input_datapoints: list[Union[InputDataEntityModel, StaticDataEntityModel]] = []
        input_datapoints.extend(input_data.input_entities)
        input_datapoints.extend(input_data.static_entities)

        input_values: dict[str, DataPointGeneral] = {}

        input_data_model = get_component_input_data_model(
            component_type=self.component_config.type
        )
        for datapoint_name, datapoint_config in self.io_model.input.__dict__.items():
            if datapoint_config is None:
                continue
            try:
                datapoint_config = IOAllocationModel.model_validate(datapoint_config)
            except ValidationError as e:
                logger.warning(
                    f"Invalid input configuration for {datapoint_name} "
                    f"in {self.component_config.id}: {e}"
                )
                continue

            datapoint = self.get_component_input(
                input_entities=input_datapoints, input_config=datapoint_config
            )

            # Skip optional datapoints with no value
            if (
                not input_data_model.model_fields[datapoint_name].is_required()
                and datapoint.value is None
            ):
                continue
            input_values[datapoint_name] = datapoint

        input_values_raw: dict[str, Any] = {}
        for key, value in input_values.items():
            input_values_raw[key] = value.model_dump()

        self.input_data = cast(
            TypeInputData, input_data_model.model_validate(input_values_raw)
        )

    def prepare_component(self):
        """
        Function to prepare the component.
        This function should be implemented in each component to prepare the component.
        """
        logger.debug(
            "Prepare component is not implemented in the base class"
            f" for {self.component_config.id}"
        )

    def calculate(self):
        """
        Function to calculate the output of the component.
        This function should be implemented in each component to calculate the output.

        The function should use the input data stored in self.input_data
        and the static data stored in self.config_data to perform the calculation.
        The result should be stored in self.output_data.

        Raises:
            ValueError: If the calculation fails due to invalid input data.
            KeyError: If a required input data point is missing.
            RuntimeError: If the calculation cannot be performed for other reasons.
        """
        logger.debug(
            "Calculate is not implemented in the base class"
            f" for {self.component_config.id}"
        )

    def run(self, data: InputDataModel) -> list[DataTransferComponentModel]:
        """
        Run the component.
        
        Args:
            data (InputDataModel): Input data for the component, \
                including all necessary entities.
        Returns:
            list[DataTransferComponentModel]: List of data transfer components.
        """
        components: list[DataTransferComponentModel] = []

        if self.io_model is None:
            logger.warning(f"IOModel of {self.component_config.id} is not set.")
            return components

        try:
            self.set_input_data(input_data=data)
        except (ValueError, KeyError) as e:
            logger.error(
                f"Setting input data failed for {self.component_config.id}: {e}"
            )
            return components

        try:
            self.calculate()
        except (ValueError, KeyError, RuntimeError) as e:
            logger.error(f"Calculation failed for {self.component_config.id}: {e}")
            return components

        try:
            output_model = get_component_output_data_model(
                component_type=self.component_config.type
            )

            if not hasattr(self, "output_data"):
                raise ValueError("Output data is not set in the component.")

            self.output_data = cast(
                TypeOutputData,
                output_model.model_validate(self.output_data.model_dump()),
            )

        except KeyError as e:
            logger.error(
                f"Getting output model failed for {self.component_config.id}: {e}"
            )
            return components
        except (ValidationError, ValueError) as e:
            logger.error(
                f"Output validation failed for {self.component_config.id}: {e}"
            )
            return components

        for datapoint_name, datapoint_config in self.io_model.output.__dict__.items():
            if datapoint_config is None:
                continue
            try:
                datapoint_config = IOAllocationModel.model_validate(datapoint_config)
                datapoint = DataPointGeneral.model_validate(
                    getattr(self.output_data, datapoint_name, None)
                )
            except ValidationError as e:
                logger.error(
                    f"Validating datapoint config failed for {datapoint_name}: {e}"
                )
                continue

            components.append(
                DataTransferComponentModel(
                    entity_id=datapoint_config.entity,
                    attribute_id=datapoint_config.attribute,
                    value=datapoint.value,
                    unit=datapoint.unit,
                    timestamp=datapoint.time or datetime.now(timezone.utc),
                )
            )

            logger.debug(
                f"Calculated {datapoint_name}: {datapoint.value} {datapoint.unit} "
                f"for {self.component_config.id}."
            )

        return components

    def calibrate(
        self,
        static_data: Optional[list[StaticDataEntityModel]] = None,
    ):
        """
        Calibrate the component
        - This function updates the static data \
            and prepares the component for operation with the new static data
        - This function can be used to adjust parameters of the component, \
            so it needs to be extended
        Args:
            static_data (Optional[list[StaticDataEntityModel]]): Static data for the component
        """
        if static_data is not None:
            try:
                logger.debug(
                    f"Reloading static data for the component {self.component_config.id}"
                )
                self.set_component_config_data(
                    static_data=static_data, static_config=self.component_config.config
                )
                self.prepare_component()
            except ComponentValidationError as e:
                logger.error(
                    f"Failed to reload static data for {self.component_config.id}: {e}"
                )
                raise



"""
Import thermal storage components and configurations.
"""

from encodapy.components.thermal_storage.thermal_storage import ThermalStorage
from encodapy.components.thermal_storage.thermal_storage_config import *



"""
Description: This module contains models for various types \
    of datapoints used in the controller component.
Author: Martin Altenburger
"""

from datetime import datetime
from typing import Any, Optional
from pydantic import BaseModel, ConfigDict
from encodapy.utils.units import DataUnits
from encodapy.utils.mediums import Medium


# Models to hold the data
class DataPointGeneral(BaseModel):
    """
    Model for datapoints of the controller component.
    
    Contains:
        value: The value of the datapoint, which can be of various types \
            (string, float, int, boolean, dictionary, list, DataFrame, or None)
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    value: Any
    unit: Optional[DataUnits] = None
    time: Optional[datetime] = None


class DataPointNumber(DataPointGeneral):
    """
    Model for datapoints of the controller component.

    Contains:
        value: The value of the datapoint, which is a number (float, int)
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: float | int


class DataPointString(DataPointGeneral):
    """
    Model for datapoints of the controller component.

    Contains:
        value: The value of the datapoint, which is a string
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: str


class DataPointDict(DataPointGeneral):
    """
    Model for datapoints of the controller component.

    Contains:
        value: The value of the datapoint, which is a dictionary
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: dict


class DataPointBool(DataPointGeneral):
    """
    Model for datapoints of the controller component.

    Contains:
        value: The value of the datapoint, which is a boolean
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: bool


class DataPointMedium(DataPointGeneral):
    """
    Model for datapoints of the controller component which define the medium.

    Contains:
        value: The value of the datapoint, which is a Medium representing the medium
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: Medium



"""
Description: This file contains the class BaererToken,\
    which is used to check if an OAuth2 bearer token is valid.
Author: Martin Altenburger
"""

from typing import Union
import requests
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session


class BaererToken:
    """
    Checks if an OAuth2 bearer token is valid.

    Returns:
        bool: True, if token is valid
    """

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        token_url: str = None,
        token: Union[str, None] = None,
    ) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.token_url = token_url
        self.token = token
        if token is not None:
            self.token_typ = "static"
        else:
            self.token_typ = "limited"

    def _is_token_valid(
        self,
    ) -> bool:
        """
        Checks if an OAuth2 bearer token is valid.

        Returns:
            bool: True, if token is valid
        """
        if self.token_typ == "static":
            return True

        response = requests.get(
            url=f"""{self.token_url}/tokeninfo""",
            params={"access_token": self.token},
            timeout=10,
        )

        if response.status_code == 200:

            token_info = response.json()
            expires_in = token_info.get("expires_in")

            if expires_in is None or expires_in > 10:
                return True

        return False

    def _get_new_token(self) -> None:
        """
        Function to get new baerer-token from oauth2-provider
        """
        client = BackendApplicationClient(client_id=self.client_id)
        oauth = OAuth2Session(client=client)
        self.token = oauth.fetch_token(
            token_url=self.token_url,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )["access_token"]

    @property
    def token_type(self) -> str:
        """
        Returns the token type
        """
        return self.token_type

    @property
    def baerer_token(self) -> str:
        """
        Returns the baerer-token
        """
        return f"""Bearer {self.token}"""

    def check_token(self) -> None:
        """
        Function to check if the actual baerer-token is valid and if not,\
            get a new baerer-token from oauth2-provider

        Returns:
            bool: True if old token is valid, false if new token has been received
        """
        if self._is_token_valid():
            return True
        self._get_new_token()
        return False
